/*
 *  Modal Logic To Sat Translator
 *
 *  2005-2006
 *  Michele Vescovi
 */
#ifndef FORMULA_H_INCLUDED
 #define FORMULA_H_INCLUDED



#include <set>

using namespace std;



#include "Def.h"
#include "Term.h"
#include "Label.h"



/*
This class stores the set of terms that constitutes a modal formula, 
in a tree or DAG structure.
This class follows the singleton patteren, so is istanceable once.
*/
class Formula
{
public:

  typedef set< Term*, TermComparator > dag_t;

  // Reset all counters (in Formula and Term)
  static void resetCounters () 
  { 
    Formula::clauses_counter = 0; 
    Term::resetCounters(); 
    Formula::resetDagCounters();
  }

  // Singelton Pattern: get the only available istance of Formula
  static Formula &getFormula ()
  {
    if (!Instance)
      Instance = new Formula();
    return (*Instance);
  }

  virtual ~Formula ();

  // It says if the formula terms are organized as tree or DAG
  bool isDag () const { return is_dag; }
  // It says if there is a root term
  bool existsRoot () const { return (Root != 0); }
  
  // Returns the root term of the formula 
  TermPtr &getRoot () const
  { assert(Root != 0); return (*Root); }
  
  // Sets the root term of the formula
  bool setRoot (Term *r, bool neg);
  bool setRoot (TermPtr &TP);

  // Deletes all terms and clears any attributes of the formula
  void ClearFormula ();
  
  // Transforms the formula (if still stored as tree) in Negative Normal Form
  bool treeToNNF ();

  // Transforms the formula (if still stored as tree) in Box Normal Form
  bool treeToBNF ();
  
  // Removes from the formula (only if still stored as tree in NNF or BNF) all the occurences of boolean values
  // (true/false are always removed from boolean subformulas, and, if requested, also BOX true and DIAMOND false can be removed)
  bool treeTrueFalseElimination (bool modal_elimination = false) 
  { 
    assert(Root != 0); 
    assert(Term::isNNF() || Term::isBNF());
    return Root->getPtr()->trueFalseElimination(Root, modal_elimination); 
  } 

  // Calls association of boolean on the whole formula (only if still stored as tree in NNF or BNF)
  bool treeBoolAssociation ()  
  { 
    assert(Root != 0);
    assert(Term::isNNF() || Term::isBNF());
    return Root->getPtr()->boolAssociation();  
  }
  
  // Calls semplification of modal operators on the whole formula (only if still stored as tree in NNF or BNF)
  bool treeModalSimplifycation ()
  { 
    assert(Root != 0); 
    assert(Term::isNNF() || Term::isBNF());
    return Root->getPtr()->modalSimplifycation(Root);  
  }

  // Converts the storing structure of the formula from a tree to a DAG
  // (stores in the passed variables some statistics about the dag construction)
  unsigned long createDAG (unsigned long *bool_match = 0, unsigned long *modal_match = 0, unsigned long *dag_size = 0)
  { 
    assert(Root != 0);
    if (is_dag) return 0;
    
    resetDagCounters ();
    
    // The construction of the DAG starts from the root but procedes recursively in an BOTTOM-UP manner        
    Root->assign( insertInDAG(Root->getPtr()), Root->isNegated() ); 
    is_dag = true; 
    
    if (dag_size) *dag_size = Term::getDagCounter();
    if (bool_match) *bool_match = Formula::bool_dag_matchings;
    if (modal_match) *modal_match = Formula::modal_dag_matchings;
      
    return (Formula::dag_matchings);
  } 

  // Sets (if "set_unset" is true) or unset (otherwise) the specified flags in all terms of the formula
  // It works only if the formula is stored as DAG
  void touchAllFlags (unsigned int flags, bool set_unset = true);

  // Translate the formula in a SAT problem 
  // It is possible to specify:
  // - the file for output (translation)
  // - the file for header information (that depends on the translation type)
  // - an option file for the mapping between identifiers used in translation and propositional-variables/states
  // - the order to follow in translation (brows the Kripke structure states in BFS or DFS order)
  // (if passed stores in "num_labels", "num_variables" and "num_clauses" some statistics about the translation)
  void translateToSat (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, bool bfs = true,
                       unsigned long *num_labels = 0, unsigned long *num_variables = 0, unsigned long *num_clauses = 0);
  
  // Produces a graph representation of the formula in the DOT format
  void printFormulaGraph (FILE *f, char *graph_name = 0);
  
protected:

  // Counter of clauses used in translation
  static unsigned long clauses_counter;

  // Statistics about the construction of DAG
  static unsigned long dag_matchings;
  static unsigned long bool_dag_matchings;
  static unsigned long modal_dag_matchings;

  static void resetDagCounters ()
  { Formula::dag_matchings = Formula::bool_dag_matchings = Formula::modal_dag_matchings = 0; }

  // It says if the formula is strored as dag or tree
  bool is_dag;

  // Points to the root term of the formula
  TermPtr *Root;
  
  // Stores all the shared terms, implemented as a std map (tree)
  dag_t DAG;
  /* 
   When terms are stored as DAG, they are maximally shared, so equality 
   between terms may be verified using a simple pointer comparison.
  */

  /*
  Inserts a term in the DAG structure, makes it shared if needed.
  Returns the pointer to the shared term, the original term may be destroyed.
  */
  Term *insertInDAG (Term *t);
  
private:

  // Static instance of formula (singleto pattern)
  static Formula *Instance;

  // Private constructor (can be called only from methods of Formula, a.g. getInstance)
  Formula ();

  // Stores the root of the Kripke structure (used in translation to SAT)
  Label *RootLabel;

  // Calls the translation of the root term of the formula
  term_transl_struct translateRoot (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, TermPtr *R, Label *l);
  
  // Calls the recursive translation of the non modal part of the subformulas that hold in the state (label) "l"
  // (at the end of the call lists Boxes and Diamonds contain all the BOX and DIAMOND operator that hold in "l")
  void translateNonModals (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l);

  // Calls the translation of a list of BOX operators in the current state (label) "l" (where no DIAMOND operators exists at the same state)
  void translateOnlyBoxes (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Boxes, Label *l);

  // Calls the translation of a DIAMOND operator in the current state (label) "l", and the list of BOX operators that hold in the same state
  // Returns the new state (label) generated from the diamond operator
  Label *translateDiamond (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, term_transl_struct Diamond, list<term_transl_struct> *Boxes, Label *l);

  // Produces the (optional) header for the translation
  void translationHeader (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, unsigned long ids_counter, unsigned long clauses_counter);
};

#endif // FORMULA_H_INCLUDED

