/*
 *  Modal Logic To Sat Translator
 *
 *  2005-2006
 *  Michele Vescovi
 */
#ifndef LABEL_H_INCLUDED
 #define LABEL_H_INCLUDED



#include <cstdio>

#include <list>
#include <string>

using namespace std;



class Term;


/* QUI */ // Spostare in Def.h ????

// Root term structure
/*
 Contains:
 - the pointer to the subformula 
 - true if the subformula has to be negated (only in BNF)
 - the cnf id assigned to the subformula in the state
*/
typedef struct 
{ 
  Term *term;
  bool negated;  
  unsigned long cnf_id; 
} 
term_transl_struct;



/*
Represents a state in the (tree) Kripke structure. 
Every state of the structure is identificated by a label, i.e. the 
order of the branchs traversed in the structure to reach the state
(sequence of numbers)
*/
class Label
{
public:

  static void resetLabelIdCounter () { Label::unique_id_counter = 0; }
  static unsigned long getLabelIdCounter () { return Label::unique_id_counter; }

  /***** Routine methods *****/

  // Constructor 
  /*
  It takes the pointer to the modal Term who generates this new state
  (diamond) and the pointer to the parent state.
  The state constructed becomes child of the parent label and gets
  the next order number among the children of the parent label.
  Construct root state on default.
  */
  Label (Term *creator = 0, Label *parent = 0);
  
  // Destructor (destroies all children too)
  virtual ~Label ();
  
  // Returns the branch order (label) of this state
  inline int getLabel () const { return label; };
  
  // Returns the number of children states in the structure
  inline int getChildrenNum () const { return Children->size(); };
  
  // Returns a reference to the list of children states
  inline list<Label *> &getChildren () const { return *Children; };

  // Returns the pointer to the modal term of the formula who has generated this state
  inline Term *getCreator () { return modal_creator; };
  // Returns the pointer to the parent state
  inline Label *getParent () { return Parent; };

  // Returns a unique id for the state (order of creation of the label)
  // ...it can be used to compare uniquely two label without using pointer values  
  inline unsigned long getId () { return unique_id; }
  
  // Returns a pointer to the textual description of the label 
  // Note: the first call (on construction) causes the allocation of the string
  const char *getText ();

  // Deallocate only this label but not its children states
  void destroyThis ();


  /*
  Define methods using ClassChunker to realiza pool allocation (and deallocation) of labels...
  
  // Allocate a new term
  static Label *allocate(Term *creator = 0, Label *parent = 0);
  
  // Deallocate a term
  static void free (Label *l);
  */
 
  // Recursively print the descrition of the labels
  // Note: without construt the string of the description 
  // Note: don't use if parent states have been destroyed with destroyThis()
  inline void print (FILE *f = stdout) const
  {  
    if (!root)
    {
      if (Parent) Parent->print(f);
      fprintf(f, "_%d", label);
    }
    else 
      fprintf(f, "%d", label);
  }

  
  /***** Methods for translation of modal fomulae *****/
  
  // Set all the root terms of the translation for this state 
  // ... i.e. the subformulae that have to hold in this state
  inline void setRootTerms (list<term_transl_struct> *RTs)
  { 
    if (RootTerms) delete RootTerms;
    RootTerms = RTs; 
  }

  // Returns all the root terms of the translation for this state 
  inline list<term_transl_struct> *getAllRootTerms ()
  { 
    list<term_transl_struct> *tmp = this->RootTerms;
    this->RootTerms = 0; 
    return tmp;
  }
  
  // Returns if is set a non empty list of root terms for the state
  inline bool existsRootTerm ()
  { 
    if (RootTerms)
      return (RootTerms->size() > 0); 
    else 
      return false;
  }
  
  // Returns and remove from the list the first root term for the state
  // (if there not exists root terms, it returns a null root term)
  term_transl_struct popRootTerm ()
  { 
    term_transl_struct T;
    T.term = 0;
    T.cnf_id = 0;
    
    if (RootTerms && (RootTerms->size() > 0))
      T = RootTerms->front();
    return T;
  }

protected: 

  // Add a child state
  // In particular sets the parent dependent attributes of the child state (label, parent, root)
  // Returns the label value of the new child. 
  int addChildLabel (Label *child);

  /*
  // Pool allocator for labels
  static ClassChunker<Label> LabelPoolAllocator;
  */

  // Unique state id, generated on construction 
  unsigned long unique_id;

  // Lable, i.e. branch order of this state with respect to the parent
  unsigned int label;


  // Children states (generated by DIAMONDs that hold the state)
  list<Label *> *Children;   
  
  // Parent label
  Label *Parent;

  // Term (DIAMOND) that has generated the state
  Term *modal_creator;

  // Is the state the root state of the tree?
  bool root;

  // Textual descrition of the label
  // (All the label values from root to here, separated by '_')
  string *text;


private:

  static unsigned long unique_id_counter;

  list<term_transl_struct> *RootTerms;
};



typedef list<Label *>::iterator label_it_t;


#endif // LABEL_H_INCLUDED
