/*
 *  Modal Logic To Sat Translator
 *
 *  2005-2006
 *  Michele Vescovi
 */
#ifndef TERM_H_INCLUDED
 #define TERM_H_INCLUDED



#define BIT(x) (1 << (x))



#include <cstdio>
#include <cstring>

#include <set>
#include <list>
#include <string>

using namespace std;



#include "Def.h"
#include "Label.h"



class TermPtr;

typedef list<TermPtr> term_list_t;
typedef list<TermPtr>::iterator term_it_t;



// Generic term of a modal formula (in the parse tree o in a DAG)
class Term
{
public:  

  friend class Formula;

    
  /***** Definitions *****/
  
  // Possible types of terms
  enum term_t 
  {
    UNKNOWN = 0,
    BOOLEAN_VALUE,  
    PROPOSITIONAL_VAR,
    BOOLEAN_OPERATOR,
    MODAL_OPERATOR
  };

  enum flags_t
  {
    F_SHARED       = BIT(1),  // True if it is a shared term (stored in a DAG)
    F_ROOT         = BIT(2),  // True if it is the root term (of the tree or DAG)
    F_VISITED      = BIT(3),  // True if the Term is visited (used in algorithms that traverse the DAG)
    F_TRANSLATED_P = BIT(4),  // True if the Term is visited (used in translation to SAT)
    F_TRANSLATED_N = BIT(5)   // True if the Term is visited (used in translation to SAT)
  };
  
  
  /***** Operations over static attributes *****/

  static void resetForm () { Term::is_NNF = Term::is_BNF = false; }
  static void resetCounters () { Term::term_id_counter = Term::dag_id_counter = Term::cnf_id_counter = 0; }
  static void resetCnfCounter () { Term::cnf_id_counter = 0; }
  static unsigned long getCnfCounter () { return Term::cnf_id_counter; }
  static unsigned long getDagCounter () { return Term::dag_id_counter; }

  static bool isNNF () { return Term::is_NNF; }
  static bool isBNF () { return Term::is_BNF; }
  
  
  /***** Standard operations over terms *****/

  // Empty (standard) constructor
  Term ();

  // Destructor
  virtual ~Term ();

  // Returns the parent term (the value is valid if the term is not shared)
  inline Term *getParent () const { return Parent; }
  
  // Returns the type of the term
  inline term_t getType () const { return type; }
  
  // Returns if the term is an operator (boolean or modal)
  inline bool isOperatorType () const { return ((type == BOOLEAN_OPERATOR) || (type == MODAL_OPERATOR)); }
  
  // Arity of the term (number of operands)
  inline unsigned int getArity () const 
  { return Children.size(); }
  
  // Adds a child term (method valid if the term is not shared)
  void addChild (TermPtr &child);
  
  /*
   Destroys the current term and transfers all the children to the parent
   (method valid if the term is not shared)
   - "it" store the position of this term in the Children list of Parent
     (reference, it is modified in a consistent way)
  */
  void joinToParent (term_it_t &it);
  
  // Default "joinToParent" call 
  // (the position in the Children list of Parent is in front)
  void joinToParent ();
  
  // Negate the term (returns if also the TermPtr have to be negated)
  inline virtual bool Negate () { return true; }
  
  // Unset all flags
  void clearFlags () { flags = 0; }
  // Set a flag
  void setFlag (unsigned int f) { flags |= f; }
  // Unset a flags
  void unsetFlag (unsigned int f) { flags &= ~f; }
  // Check a flag value
  bool checkFlag (unsigned int f) const { return (flags & f) != 0; }
  // Returns all flags
  unsigned int getFlags () const { return flags; }

  // Set the SHARED falg and assign an unique id for the term in the DAG
  void share () { setFlag(Term::F_SHARED); id = ++Term::dag_id_counter; }

  // Operator less used when comparing terms in DAG
  bool operator< (const Term &T) const;


  
  /***** Recursive transformation operations on terms (when terms are in a tree) *****/
  
  // Recursively propagates the transformation of the tree in Negative Normal Form (negations only on propositional varibles)
  // It can change the parity value of the term (strored in TermPtr)...
  virtual void NNF (TermPtr *myPtr); 

  // Recursively propagates the transformation of the tree in Box Normal Form (negations only on box operators or propositional varibles)
  // It can change the parity value of the term (strored in TermPtr)...
  virtual void BNF (TermPtr *myPtr); 

  
  // Recursively propagates the elimination of boolean values (true-false)
  // (elimination is done in boolean subformulas, and, if modal_elimintation is true, also in modal subformulas)
  // Returns true if leads to changes, it can be change the root term
  virtual bool trueFalseElimination (TermPtr *root, bool modal_elimination = false); 
 
  // Recursively propagates the association of equivalent boolean operators 
  // Returns true if leads to changes
  virtual bool boolAssociation (); 
  
  // Recursively propagates the semplification of modal and boolean operators 
  // Returns true if leads to changes, it can be change the root term
  virtual bool modalSimplifycation (TermPtr *root); 

  // Deallocate only this terms and all its unshared terms
  void destroy();


  /***** Methods used in translation to SAT *****/  
  
  /*
    Return the CNF id to be used in translation of the term in the state (label) "l".
    If "use_id" is set, sets the value of the last id and label used in translation
    ...otherwise returns the last used id, if "l" is equal to the last label, otherwise generate a new id.
    Possibly call the print of the mapping between id and variables.
    (Virtual... every inheritor class could implement with respect to the term type) 
  */
  virtual unsigned long getCnfId (Label *l, unsigned long use_id = 0, FILE *map_f = 0, transl_t translation = T_UNKNOWN);

  // Print the SAT translation for the root term (that have to be satisfiable)
  // Returns the number of printed clauses 
  int translateAsRoot (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, Label *l, bool negated);
  
  // Print the SAT translation for any non modal term.
  // The modal terms reached (Diamonds and Boxs) are pushed in the respective lists with their CNF ids.
  // Returns the number of printed clauses.
  virtual int translateNonModal (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l, bool negated);


  /***** Methods for debug prints of terms *****/

  // Prints the term id 
  void printDotId (FILE *f) const { fprintf(f, "%d", id); }

  // Recursively prints the translation (in DOT format) of the this term
  void printDotGraph (FILE *f);

protected:

  // Counter used to assign ids to terms during the construction of the parse tree
  static unsigned long term_id_counter;
  // Counter used to assign ids to terms during the construction of the dag
  static unsigned long dag_id_counter;
  // Counter used to assign ids to term in states during the translation to SAT
  static unsigned long cnf_id_counter;

  // True if the root term has been transformed in NNF
  static bool is_NNF;
  // True if the root term has been transformed in BNF
  static bool is_BNF;


  /***** Term Attributes *****/

  // Constructor with a child (used for MODAL terms)
  Term (TermPtr &child);
  // Constructor with two children (used for binary operator terms)
  Term (TermPtr &left, TermPtr &right);
  // Constructor with a list of children (used for BOOLEAN terms)
  Term (term_list_t *children);

  // Used to store the flag values
  unsigned int flags;


  // Type of the term
  term_t type;

  // Pointer to the parent node in the parse tree (no more valid in DAG)
  Term *Parent;
  
  // List of (TermPtr pointer + negation) the children term
  term_list_t Children;


  // Unique id assigned to the term 
  // (a progressive id is assigned in construction of the parse tree and of the DAG)
  unsigned long id;
  
  // Last CNF id assigned to the term (+ state) during the translation
  unsigned long last_transl_id;
  // Id of the label (state) in which the CNF id of the term is valid
  unsigned long last_transl_label_id;

  
  /***** Methods on tree terms (not valid if term is shared) *****/  
 
  // Transform the term in Negative Normal Form 
  // (Virtual... every inheritor class could implement with respect to the term type) 
  inline virtual void pushNegationInwards (TermPtr *myPtr) {}; 

  // Transform the term in Box Normal Form 
  // (Virtual... every inheritor class could implement with respect to the term type) 
  inline virtual void toBoxNormalForm (TermPtr *myPtr) {}; 
    
  // Eliminates boolean values from term (only in NNF or BNF)
  // Returns true if the leads to modifications
  // (Virtual... every inheritor class could implement with respect to the term type)   
  inline virtual bool eliminateTrueFalse () { return false; }; 
    
  // Does the possible boolean associations in term (only in NNF or BNF)
  // Returns true if the leads to modifications
  // (Virtual... every inheritor class could implement with respect to the term type)   
  inline virtual bool flatBooleanOp () { return false; }; 
 
  // Does the possible modal simplifications in term (only in NNF or BNF)
  // Returns true if the leads to modifications
  // (Virtual... every inheritor class could implement with respect to the term type)   
  inline virtual bool simplifyModalOp () { return false; };     

  
  /***** Methods used in construction of DAG *****/  
  
  // Used in operator< to compare terms of the same type and arity 
  // Returns 0 if equale, < 0 if this term is less than T, > 0 otherwise
  // (Virtual... every inheritor class could implement with respect to the term type) 
  virtual int compare (const Term *T) const { return 0; };  

  // Returns if, during DAG constructions, repeated children can be removed
  inline virtual bool dropRepeatedChildren () const { return true; };  


  /***** Methods for debug prints of terms *****/  
  
  // Prints the term attributes in DOT format 
  // (Virtual... every inheritor class could implement with respect to the term type) 
  inline virtual void printDotAttributes (FILE *f) const {;}
    
private:
  
  // Inits the attributes of the term
  void init ();

  // If the root term is modal it can be repeatly semplificated
  // To be called after the conversion of tree in NNF (and not if terms are shared)
  void eliminateModalRoot (TermPtr *root);
};



// Stores the pointer to a term (with optional negation)
class TermPtr  
{ 
public:

  friend class Formula;

  // Constructor
  TermPtr (Term *p, bool n = false) 
  { 
    assert(p != 0); 
    ptr = p;
    negated = false;     
    if (n) Negate(); 
  };

  // Copy constructor
  TermPtr (const TermPtr &T) 
  { (*this) = T; };
  
  // Destructor
  ~TermPtr () { };

  // Returns the term pointer
  inline Term *getPtr() { return ptr; };
  // Returns the term pointer (const)
  inline const Term *getConstPtr() const { return (const Term*)ptr; };
  
  // Returns if the term is negated
  inline bool isNegated () const { return negated; };
  // Negate the term
  // Calls the negation methos of the term, and update "negate" if necessary
  inline void Negate () { if (ptr->Negate()) negated = !negated; };

  // Assignment operator
  inline TermPtr &operator= (const TermPtr &T) { ptr = T.ptr; negated = T.negated; return (*this); };
  // Assignment using params
  inline void assign (Term *p, bool n) { assert(p != 0); ptr = p; negated = n; };

  
  // Prints in DOT format of the term 
  void printDotGraph (FILE *f) const
  {
    // Prints the link from parent to this term
    fprintf(f, " -> ");
    ptr->printDotId(f);
    // If the term is negate the information is put on the label of the link
    if (negated)
      fprintf(f, " [label=\"NOT\"]");
    fprintf(f, ";\n");
    // The call the print of the term
    ptr->printDotGraph(f);
  }

private:  

  // Pointer to the term
  Term *ptr; 

  // Store if the term is negated
  bool negated;

  // Empty constructor 
  TermPtr () { ptr = 0; negated = false; };
};





// Term that represents a boolean constant value
class BooleanValue: public Term 
{
public:  

  // Constructor
  BooleanValue (bool v);
  
  // Returns the boolena value (true or false)
  inline bool getValue () const { return value; }

  // Negation change the value of the boolean constant (these terms are never negated) 
  inline virtual bool Negate () { value = !value; return false; }

  // Return the CNF id to be used in translation of the term 
  // The CNF id used is unque, indipendently from the state (label)
  virtual unsigned long getCnfId (Label *l, unsigned long use_id = 0, FILE *map_f = 0, transl_t translation = T_UNKNOWN);

  // Translate the boolean term to SAT
  // Returns the number of printed clauses.
  int translateBooleanValue (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation);
  
  // Prints the correspondence between the boolean value and the CNF id used in translation
  int printVarMap (FILE *map_f, transl_t translation, Label *l);
  
protected:
  
  // Store the boolean value of the node
  bool value;  


  // Compare two boolean values (false is less than true)
  virtual int compare (const Term *T) const
  { return ( (int)value - (int)(((BooleanValue*)T)->getValue()) ); };  

  // Prints the boolean value of the term in DOT format
  inline virtual void printDotAttributes (FILE *f) const  
  { fprintf(f, " [label=\"%s\"]", (value ? "true" : "false") ); }   
  
private:
};



// Term that represents a propositional variable
class PropositionalVar: public Term 
{
public:  

  // Constructor
  PropositionalVar (char *variable_name);
  // Copy Constructor
  PropositionalVar (PropositionalVar &PV);

  /* virtual int translateNonModal (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l, bool negated); */
 
  // Prints the correspondence between the variable and the CNF id used in translation in the given state (label)
  int printVarMap (FILE *map_f, transl_t translation, Label *l);
  
protected:

  // Name of this propositional variable
  string Name;  

  /*QUI*/ // Introdurre un identificatore di variabile da 1 in su
          // da usare sia per fare una compare piu' veloce sia per la unit propagation   

  // Compare two propositional variables (alphanumeric compare on variable names is used)
  virtual int compare (const Term *T) const
  { return ( strcmp(Name.c_str(), ((PropositionalVar*)T)->Name.c_str()) ); };
  
  // Prints the variable name in DOT format
  inline virtual void printDotAttributes (FILE *f) const  
  { fprintf(f, " [label=\"%s\"]", Name.c_str()); }   

private:
};



// Term that represents a boolean operator (and, or)
class BooleanOp: public Term
{
public:  

  // Constructor (n-ary operator case): needs the list of operands terms and the type of the boolean operator
  BooleanOp (term_list_t *operands, operator_t op);
  // Constructor (binary operator case): needs the two operands terms and the type of the boolean operator
  BooleanOp (TermPtr &left, TermPtr &right, operator_t op);

  // Returns the boolean operator
  inline operator_t getOperator () const { return bool_op; }

  // Recursively call the SAT translation for any non modal term.
  // Returns the number of printed clauses.
  virtual int translateNonModal (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l, bool negated);
  
protected:

  // Type of the operator
  operator_t bool_op;

  // Compare two booelan operators (the operator id is used)
  virtual int compare (const Term *T) const
  { return ( (int)bool_op - (int)(((BooleanOp*)T)->getOperator()) ); };  

  // Negative Normal Form transformation of this term 
  // (if negated, negates all children and swap operator AND<->OR)
  virtual void pushNegationInwards (TermPtr *myPtr);

  // Transform the term in Box Normal Form (for boolean operators equivalent to NNF)
  inline virtual void toBoxNormalForm (TermPtr *myPtr) 
  { pushNegationInwards(myPtr); }
  
  // Eliminates true and false valued among the children terms
  // (TRUE-in-AND and FALSE-in_OR are eliminated, FALSE-in-AND = FALSE, TRUE-in-OR = OR)
  virtual bool eliminateTrueFalse (); 
  // Associates equal boolean operator between this term and children
  // (AND-of-ANDs = AND  and  OR-of-ORs = OR)
  virtual bool flatBooleanOp (); 
  // Simplyfies equal modal operators accordingly with this boolean operator
  // (AND-of-BOXes = BOX-of-AND  and OR-of-DIAMONDs = DIAMOND-of-OR)
  virtual bool simplifyModalOp ();     

  // Prints the operator type in DOT format
  inline virtual void printDotAttributes (FILE *f) const  
  { 
    fprintf(f, " [label=\"");
    switch (bool_op)
    {
      case O_AND: fprintf(f, "AND"); break;
      case O_OR: fprintf(f, "OR"); break;
      default: break;
    }
    fprintf(f, "\"]");
  }

  // Translate to SAT this boolean operator term
  // Returns the number of printed clauses.
  int translateBoolOp (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, Label *l, bool negated);
  
private:

  // Change the order of the children to permit the modal simplyfication
  // ...modal children nodes with the correct operator and polarity are positioned in front of the list.
  // (Correct operator is DIAMOND for OR and BOX for AND)
  // Returns the position in children list of the first non correct or modal child.
  // In "swapped" stores the number of correct modal operators found.
  // In "contains_boolean" if modal operators found have boolean children
  term_it_t sortModalChildren (operator_t modal_op, bool polarity, int *swapped, bool *contains_boolean);
};



// Term that represents a modal operator (box, diamond)
class ModalOp: public Term
{
public:  

  // Constructor: needs operand term and the type of the modal operator
  ModalOp (TermPtr &sub_formula, operator_t op);

  // Returns the boolean operator
  inline operator_t getOperator () const { return modal_op; }

  // Returns the reference to the only child
  TermPtr &getChild () { return Children.front(); } 

  // Since these terms are not modal no translation are done, but this term are push in the list of diamonds or boxes
  virtual int translateNonModal (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l, bool negated);
  
  // Print the SAT translation of a box term with respect to a given diamond (and the respective new state)
  // Returns the number of printed clauses.
  int translateBox (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, Label *l, unsigned long use_id, Label *child_l = 0, unsigned long id_diamond = 0);

  // Print the SAT translation of a diamond term in the given new state
  // Returns the number of printed clauses.
  int translateDiamond (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, Label *l, unsigned long use_id, Label *child_l);
  
protected:

  // Modal operator
  operator_t modal_op;

  // Compare two modal operators (the operator id is used)
  virtual int compare (const Term *T) const
  { return ( (int)modal_op - (int)(((ModalOp*)T)->getOperator()) ); };  

  // Negative Normal Form transformation of this term
  // (if negated, negates the child and swap operator BOX<->DIAMOND)
  virtual void pushNegationInwards (TermPtr *myPtr);

  // Transform the term in Box Normal Form 
  // (no modifications to BOXes and negated BOXes... DIAMONDs are translated to BOXes with the opportune negations)
  inline virtual void toBoxNormalForm (TermPtr *myPtr); 
  
  // Prints the operator type in DOT format
  inline virtual void printDotAttributes (FILE *f) const  
  { 
    fprintf(f, " [label=\"");
    switch (modal_op)
    {
      case O_BOX: fprintf(f, "BOX"); break;
      case O_DIAMOND: fprintf(f, "DIA"); break;
      default: break;
    }
    fprintf(f, "\"]");
  }
  
private:
};



// Class parameter for the standard libraries template structures (like set, map, etc.)
struct TermComparator : binary_function<Term *, Term *, bool>
{
   bool operator() (const Term *T1, const Term *T2) const
   { return ((*T1) < (*T2)); }
};

#endif // TERM_H_INCLUDED
