/*
 *  Modal Logic To Sat Translator
 *
 *  2005-2006
 *  Michele Vescovi
 */
#include "Formula.h"



#include <cstdio>



unsigned long Formula::clauses_counter = 0;

unsigned long Formula::dag_matchings = 0;
unsigned long Formula::bool_dag_matchings = 0;
unsigned long Formula::modal_dag_matchings = 0;



Formula *Formula::Instance = 0;



Formula::Formula ():
  DAG() // Construct the set of terms used for the DAG
{
  Root = 0;
  is_dag = false;   // default (on parsing) the formula is structured as tree
  
  RootLabel = 0;
}

Formula::~Formula ()
{
  ClearFormula();
}


void Formula::ClearFormula ()
{
  if (Root)
  {
    if (is_dag)
    {
      // Destroy all the term of the DAG (they are shered so is called "destroy" instead of the destructor)
      for (dag_t::iterator dag_it = DAG.begin(); dag_it != DAG.end(); ++dag_it)
        (*dag_it)->destroy();
      DAG.clear();
    }
    else
      // Deltion procedes recursively from root to leaves
      delete (Root->getPtr());
    
    delete(Root);
    Root = 0;
  }
  is_dag = false;    
  
  if (RootLabel) 
  {
    delete (RootLabel);
    RootLabel = 0;
  }
  
  Formula::resetCounters();
  Term::resetForm();
}

Term *Formula::insertInDAG (Term *t)
{
  term_it_t it, sort_it; 
  
  assert (t != NULL);

  // Check if t is already SHARED (already insert in DAG). This is a speed hint
  if (t->checkFlag(Term::F_SHARED))
    return t;

  if (t->getArity() > 1)
  {
    term_list_t SortedChildren; 
    
    // First insert and share children...
    for (it = t->Children.begin(); it != t->Children.end(); ++it)
    {
      // Insert the child in DAG and get the new pointer
      Term *dag_ptr = insertInDAG(it->getPtr());

      // Children have to be sorted by pointer to speed up the test of equivalence between two terms
      sort_it = SortedChildren.begin();
      while ( (sort_it != SortedChildren.end()) && (dag_ptr > sort_it->getPtr()) )
        ++sort_it;
      // ...now sort_it points to the first pointer equal or greather than dag_ptr
      
      if ( (! t->dropRepeatedChildren() )      ||       // if is not allowed the elimination of repeated children
           ( sort_it == SortedChildren.end() ) ||       // ...or if it is reached the end of the list of children
           ( dag_ptr != sort_it->getPtr() )    ||       // ...or if the two children are not equal
           ( it->isNegated() != sort_it->isNegated() )  // ...or if they have different parity
         )
      {        
        // ...the current child is inserted in the correct position
        TermPtr Tmp(dag_ptr, it->isNegated());          
        SortedChildren.insert(sort_it, Tmp);
      }
      // ...otherwise it is skipped (it can occour that a boolean operator becomes unary)
    }
    // Clear the old children and move the new ones in the sorted list
    t->Children.clear();
    t->Children.splice(t->Children.begin(), SortedChildren);
  }
  else 
    // The term has only a child... compute and assign the new pointer (inserting the child in the DAG)
    if (t->getArity() == 1)
      (t->Children.begin())->assign(insertInDAG((t->Children.begin())->getPtr()), (t->Children.begin())->isNegated());

  // Insertion of this term...    
  dag_t::iterator dag_it = DAG.find(t);
  if (dag_it == DAG.end())
  {
    // Term not found... this is a new node od the DAG
    DAG.insert(t);
    t->share();   // Set the term shared and return
    return t;
  }
  else
  {
    // This is a kwown node of the DAG. Destroy original.
    ++Formula::dag_matchings;
    if (t->getType() == Term::MODAL_OPERATOR) ++Formula::modal_dag_matchings;
    if (t->getType() == Term::BOOLEAN_OPERATOR) ++Formula::bool_dag_matchings;
    
    t->destroy();
    return (*dag_it);
  }
}

bool Formula::setRoot (Term *r, bool neg)
{ 
  if (Root) 
    return false;
  
  r->setFlag(Term::F_ROOT);
  Root = new TermPtr(r, neg); 
  return true;
}

bool Formula::setRoot (TermPtr &TP)
{ 
  if (Root) 
    return false;
    
  TP.getPtr()->setFlag(Term::F_ROOT);
  Root = new TermPtr(TP); 
  return true;
}

bool Formula::treeToNNF ()
{ 
  assert(Root != 0);
    
  if (!is_dag)
  {
    // If formula is still stored as tree call NNF transformation to the root term
    Root->getPtr()->NNF(Root); 
    return true;    
  }
  else 
    return false;
}

bool Formula::treeToBNF ()
{ 
  assert(Root != 0);
    
  if (!is_dag)
  {
    // If formula is still stored as tree call BNF transformation to the root term
    Root->getPtr()->BNF(Root); 
    return true;
  }
  else 
    return false;
} 

void Formula::touchAllFlags (unsigned int flags, bool set)
{
   dag_t::iterator it;
  
   if (set)
     for (it = DAG.begin(); it != DAG.end(); ++it)
       (*it)->setFlag(flags);
   else
     for (it = DAG.begin(); it != DAG.end(); ++it)
       (*it)->unsetFlag(flags);
}


void Formula::printFormulaGraph (FILE *f, char *graph_name)
{
  // Print the graph header in DOT format
  fprintf(f, "digraph %s \n{\n", graph_name ? graph_name : "Formula");
  
  if (Root) 
  {
    if (graph_name) 
      fprintf(f, "  graph [label=\"%s\"];\n", graph_name);
 
    fprintf(f, "  graph [center=true];\n");
    fprintf(f, "  edge [decorate=true];\n\n");
    
    fprintf(f, "  root [label=\"\", height=0.1, width=0.1, style=invis];\n");    
    fprintf(f, "  root");

    // If formula is stored as DAG the flag F_VISITED is used to avoid double printing of the same subformula
    if (is_dag) touchAllFlags (Term::F_VISITED, false);
    
    Root->printDotGraph(f);
  }
  fprintf(f, "\n}\n\n");
}


void Formula::translateToSat (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, bool bfs, 
                              unsigned long *num_labels, unsigned long *num_variables, unsigned long *num_clauses)
{
  // Here are stored the root terms for the current label
  list<term_transl_struct> *root_terms;
  term_transl_struct T, D;
  
  // Here are stored the Diamond (in NNF, or ~Box in BNF) and Box terms found int the current label
  list<term_transl_struct> Diamonds, Boxes; 

  // List of all the generated and not still analyzed labels
  list<Label *> L;
  list<Label *>::iterator front_it;

  // Current label
  Label *Current;

  assert(out_f);
  assert(Root != 0);
  assert(Term::isNNF() || Term::isBNF());
  // assert(is_dag);


  Formula::clauses_counter = 0;
  Term::resetCnfCounter();
  
  // Generate the root label (where the whole formula must be valid)
  RootLabel = Current = new Label(Root->getPtr()); 

  // Impose the the root term to be satisfiable
  T = translateRoot(out_f, header_f, map_f, translation, Root, Current);

  // For the root label the only root term is the formula's root term
  root_terms = new list<term_transl_struct>();
  root_terms->push_front(T);
  Current->setRootTerms(root_terms);

  // Translation label by label doesn't involve the storing of the id used for every term and label...
  // ...moreover assures (just using a flag) that the same term is not translated twice for the same label!

  // Start from the root label and iterate until unanalyzed labels exist
  L.clear();
  L.push_front(Current);
  while ( ! L.empty() )
  {
    // Extract the current label to analyze
    Current = L.front();
    L.pop_front();
    
    // Recursively translate the non modal part reachable from the current root terms
    translateNonModals(out_f, header_f, map_f, translation, &Diamonds, &Boxes, Current); 
    
    if (Diamonds.empty())
      // If no diamonds are reached boxes are aalways true... translate in this way all reached boxes
      translateOnlyBoxes(out_f, header_f, map_f, translation, &Boxes, Current);
    else 
    {
      // "front_it" store the old begin of the label list if DFS order is used
      front_it = L.begin();
      
      // For every diamond found in the current label...
      while (! Diamonds.empty())
      {
        Label *new_child_label = 0;
        
        D = Diamonds.front(); 
        Diamonds.pop_front();
       
        // A new child label is generated from the tranlation of the diamonds
        // (also every box is translated in with the current diamond
        // ...this assures that the CNF id generated and used in child terms are always the same)
        new_child_label = translateDiamond(out_f, header_f, map_f, translation, D, &Boxes, Current);

        // If the new label contains (complex - operator) root terms 
        if (new_child_label->existsRootTerm())
        {
          // ...it needs to be translated
          if (bfs)
            // BFS: list L is used as a queue 
            L.push_back(new_child_label);
          else        
            // DFS: list L is used as a stack 
            // (inserting in "front_it" assures that new labels follows in front of L the order of diamonds in Diamonds)
            L.insert(front_it, new_child_label);
        }
        else
        {
          // ...otherwise it can be destroyed since doesn't need translation
          new_child_label->destroyThis();
          new_child_label = 0;
        }
      }
    }
    // Clear the boxes list for the next iteration
    Boxes.clear();
    
    // Destroy (!!! don't use delete that free also children !!!) the current label...that will never be consider anymore
    Current->destroyThis();
  }

  // Eventually prints information in the translation header...
  if (header_f) translationHeader(out_f, header_f, map_f, translation, Term::getCnfCounter(), Formula::clauses_counter);
  
  /* It is not necessary to destroy the label tree ( delete (RootLabel) )...
     ...since "destroyThis" is called for every created label */
  RootLabel = 0;
    
  if (num_labels) *num_labels = Label::getLabelIdCounter();
  if (num_variables) *num_variables = Term::getCnfCounter();
  if (num_clauses) *num_clauses = Formula::clauses_counter;
}

void Formula::translationHeader (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, unsigned long ids_counter, unsigned long clauses_counter)
{
  if (translation == T_CNF_DIMACS)
    fprintf(header_f, "p cnf %d %d\n", ids_counter, clauses_counter); 
}

term_transl_struct Formula::translateRoot (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, TermPtr *RootPtr, Label *l)
{
  term_transl_struct T;
  Term *R = RootPtr->getPtr();
  
  Formula::clauses_counter += R->translateAsRoot(out_f, header_f, map_f, translation, l, RootPtr->isNegated());

  // Return the root term for the root label
  T.term = R;
  T.cnf_id = R->getCnfId(l);
  T.negated = Term::isBNF() && (RootPtr->isNegated());
  
  return T;
}

void Formula::translateNonModals (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l)
{
  list<term_transl_struct>::iterator rt;
  list<term_transl_struct> *r_terms = 0;
  term_transl_struct T;

  // Extracts the root terms for the current label 
  r_terms = l->getAllRootTerms();
  if (r_terms)
  {
    // First set the label and the respective id for every root term... 
    // (this avoids overwriting of the flags F_TRANSLATED_*)
    for (rt = r_terms->begin(); rt != r_terms->end(); ++rt)
      rt->term->getCnfId(l, rt->cnf_id); 

    // ...then translate each root term... 
    // ...so if a root term is reached inside the translation of another root term it isnt translated two times.
    for (rt = r_terms->begin(); rt != r_terms->end(); ++rt)
      if (rt->term->isOperatorType())
        Formula::clauses_counter += rt->term->translateNonModal(out_f, header_f, map_f, translation, Diamonds, Boxes, l, rt->negated);

    delete (r_terms);
  }
}

void Formula::translateOnlyBoxes (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Boxes, Label *l)
{
  list<term_transl_struct>::iterator b;
  
  for (b = Boxes->begin(); b != Boxes->end(); ++b)
    // Translate every box without diamond id and child label since there is no diamonds at the current label
    Formula::clauses_counter += ((ModalOp *)(b->term))->translateBox(out_f, header_f, map_f, translation, l, b->cnf_id, 0, 0); 
}

Label *Formula::translateDiamond (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, term_transl_struct Diamond, list<term_transl_struct> *Boxes, Label *l)
{
  list<term_transl_struct>::iterator b;
  term_transl_struct T;
  Term *child;
 
  // Stores if there are operator terms (that need translation) between the root terms of the new child label
  bool operator_type = false;
  
  // Stores the root terms for the new child label    
  list<term_transl_struct> *RTs = new list<term_transl_struct>();

  // The diamond genereate a child label of the current label
  Label *child_l = new Label(Diamond.term, l);
  
  // Produces the translation of the diamond    
  Formula::clauses_counter += ((ModalOp *)Diamond.term)->translateDiamond(out_f, header_f, map_f, translation, l, Diamond.cnf_id, child_l); 
  child = ((ModalOp *)Diamond.term)->getChild().getPtr();

  // Boolean have not to be included as root terms... they are "label indipendent" and don't need to be translated anymore
  // ...differently from propositional variables that have to be included as root terms to save their id in the new child label
  if (child->getType() != Term::BOOLEAN_VALUE)
  {
    // Save the root term for the new child label (term pointer and CNF id)
    T.term = child;
    T.cnf_id = child->getCnfId(child_l);
    // Used only in BNF: every "Diamond" is really a ~BOX, and its child will change polarity in translation
    // (If the form is not BNF (i.e. is NNF) negated is always false)
    T.negated = Term::isBNF() && (! ((ModalOp *)Diamond.term)->getChild().isNegated());

    RTs->push_front(T);
    
    operator_type = operator_type || (child->isOperatorType());
  }
  
  // For every box operator in the current label....   
  for (b = Boxes->begin(); b != Boxes->end(); ++b)
  {
    // ...Produces the translation of the box with respect to the current diamond and the new child label
    int translated = ((ModalOp *)(b->term))->translateBox(out_f, header_f, map_f, translation, l, b->cnf_id, child_l, Diamond.cnf_id);
    Formula::clauses_counter += translated;

    child = ((ModalOp *)(b->term))->getChild().getPtr();

    // Note: no equal boxes can occour in the list (a box is pushed in list only the first time using flag F_TRANSLATED)
    // Note: the child of the box can not be child of another box (if the formula is structured as DAG the two boxes are the same term)
    // Note: the child of the box can be the same child of the diamond but it will anyway transladet once
    
    // Boolean have not to be included as root terms... they are label indipendent and don't need to be translated anymore
    // ...differently from propositional variables that have to be included as root terms to save their id in the new child label
    if (translated && (child->getType() != Term::BOOLEAN_VALUE))
    {
      // Save the root term for the new child label (term pointer and CNF id)
      T.term = child;
      T.cnf_id = child->getCnfId(child_l);
      // Used only in BNF: every "Box" is really a BOX, and its child will conserve its polarity in translation
      // (If the form is not BNF (i.e. is NNF) negated is always false)
      T.negated = Term::isBNF() && ((ModalOp *)b->term)->getChild().isNegated();

      RTs->push_back(T);

      operator_type = operator_type || (child->isOperatorType());
    }
  }

  // If no operators are among the root terms, root term can be cleared...
  // ...since the new child label doesn't need to be translated.
  if (! operator_type) 
    RTs->clear();
  
  if (RTs->empty()) 
  {
    // If there not exist root terms free memory
    delete (RTs);
    RTs = 0;
  }

  child_l->setRootTerms(RTs);

  return child_l;
}
