/*
 *  Modal Logic To Sat Translator
 *
 *  2005-2006
 *  Michele Vescovi
 */
#include "Term.h"



#include <cstdio>



// Initialization of static fields
unsigned long Term::term_id_counter = 0;
unsigned long Term::dag_id_counter = 0;
unsigned long Term::cnf_id_counter = 0;

bool Term::is_NNF = false;
bool Term::is_BNF = false;



Term::Term ()
{
  init();
}

Term::Term (TermPtr &child)
{
  init();
  addChild(child);
}

Term::Term (TermPtr &childL, TermPtr &childR)
{
  init();
  addChild(childR);   // 'Cause addChild push children in front of the list
  addChild(childL);
}

Term::Term (term_list_t *children):
  Children()
{
  assert(children != 0);
  
  init();
  // Move children from the passed list, end free it
  Children.splice(Children.end(), *children);
  delete (children);
  
  // Set this as Parent of children terms
  for (term_it_t it = Children.begin(); it != Children.end(); ++it)
    it->getPtr()->Parent = this;
}

Term::~Term ()
{
  if (Children.size() > 0)
  {
    term_it_t it;
   
    // It will destroy recursively the tree rooted in this node
    for (it = Children.begin(); it != Children.end(); ++it)
      delete (it->getPtr());
    Children.clear();
  }
}

void Term::destroy()
{
  assert (checkFlag(F_SHARED) == false);
  
  if (Children.size() > 0)
  {
    term_it_t it;
   
    // It will recursively freen the non-shared part of the tree rooted in this node
    for (it = Children.begin(); it != Children.end(); ++it)
      if (! it->getPtr()->checkFlag(F_SHARED))
        delete (it->getPtr());
    // Clear the children list to avoid deletion inside the distructor
    Children.clear();
  }
  delete (this);
}

void Term::init () 
{
  // Standard initialization of semantic fields
  type = UNKNOWN;
  clearFlags();
  
  // Tree structure fields
  Parent = 0;

  // unique id (oreder of construction in the parse tree)
  id = ++Term::term_id_counter;

  // Fields used in translation to SAT
  last_transl_id = 0;
  last_transl_label_id = 0;
}

void Term::addChild (TermPtr &child)
{
  assert(checkFlag(F_SHARED) == false);
  
  // New child is pushed in front 
  Children.push_front(child);
  // Set this as Parent of the new child node
  if (child.getPtr()->checkFlag(F_SHARED) == false)
    child.getPtr()->Parent = this;
}

// Compare this with another term T... 
// Used to construct a DAG of the terms and verify if two terms are identical
// (on equal term compares childrens using the pointer values... childrend has to be previously ordered by this criterion)
bool Term::operator< (const Term &T) const 
{
  // Check type
  if (getType() < T.getType())
    return true;
  else if (getType() == T.getType())  
  {
    // Check arity
    if (getArity() < T.getArity())
      return true;
    else if (getArity() == T.getArity())
    {
      int cmp;
      
      // Compare terms of the same type...
      if (cmp = compare(&T)) return (cmp < 0);

      // ...this and T are equal terms of the same type... compare subterms
      
      // Recursive checking of subterms:
      // - Use pointers in comparing enhances the performances, but Children has to be sorted in this way.
      // - This works since Children are inserted in DAG first, and are sorted using the pointer value.
      term_list_t::iterator<TermPtr, const TermPtr&, const TermPtr*> it_this = this->Children.begin();
      term_list_t::iterator<TermPtr, const TermPtr&, const TermPtr*> it_T = T.Children.begin();
      for ( ; it_this != Children.end(); ++it_this, ++it_T)
        if ( it_this->getConstPtr() != it_T->getConstPtr() )
          return ( it_this->getConstPtr() < it_T->getConstPtr() );
        // Equal children can have different parity... 
        else if ( it_this->isNegated() != it_T->isNegated() )
          return ( it_T->isNegated() );
    }
  }
  // ...they are the same term
  return false;
}

// It can change the parity value of the term....
void Term::NNF (TermPtr *myPtr) 
{
  assert(myPtr != 0);

  if (checkFlag(F_ROOT))
  {
    Term::is_NNF = true;
    Term::is_BNF = false;
  }
  
  // Call to the LOCAL and VIRTUAL nnf transformation of the term
  pushNegationInwards(myPtr);
  
  // Recursive propagation of the nnf transoformation (TOP-DOWN)
  for (term_it_t it = Children.begin(); it != Children.end(); ++it)
    it->getPtr()->NNF(&(*it));
} 

// It can change the parity value of the term....
void Term::BNF (TermPtr *myPtr) 
{
  assert(myPtr != 0);
  
  if (checkFlag(F_ROOT))
  {
    Term::is_BNF = true;
    Term::is_NNF = false;
  }
  
  // Call to the LOCAL and VIRTUAL bnf transformation of the term
  toBoxNormalForm(myPtr);
  
  // Recursive propagation of the bnf transoformation (TOP-DOWN)
  for (term_it_t it = Children.begin(); it != Children.end(); ++it)
    it->getPtr()->BNF(&(*it));
} 

bool Term::trueFalseElimination (TermPtr *root, bool modal_elimination) 
{
  // "ret" is set to true if any of the association procedure returns true (there is a change)
  bool ret = false;

  assert(checkFlag(F_SHARED) == false);
  
  // Recursive propagation of the elimination procedure (BOTTOM-UP)
    for (term_it_t it = Children.begin(); it != Children.end(); ++it)
    {
      Term *T = it->getPtr();
      
      // Cannot excange the order of || operands... 
      // otherwise "trueFalseElimination" could not be recursively propagated
      ret = ( T->trueFalseElimination(0, modal_elimination) ) || ret;
      // ...passing first parameter as zero since children can not be root
      
      // If elimination causes an ennary boolean operator child to become unary...
      // ...it is removed and its child takes its place between the children of this term.
      if (T->getType() == BOOLEAN_OPERATOR) 
      {
        if (T->getArity() == 1) 
        {
          if (it->isNegated())
            ((T->Children).begin())->Negate();
          T->joinToParent(it);
        }
      }
      else 
      {
        // If is permitted the elimination of boolean values in modal operators...
        if (modal_elimination)
          if (T->getType() == MODAL_OPERATOR) 
          {
            Term *C = ((ModalOp*)T)->getChild().getPtr();
            if (C->getType() == BOOLEAN_VALUE)
            {
              if (Term::is_NNF)
              {
                // ...trasforms (BOX true) and (DIAMONDS false) respectively in (true) and (false)
                if ( ((BooleanValue*)C)->getValue() == ( ((ModalOp*)T)->getOperator() == O_BOX ) )
                  T->joinToParent(it);
              }
              else if (Term::is_BNF)
              {
                // Operator is BOX since the subformula is in BNF... 
                // ...trasforms (BOX true) and (~BOX true) respectively in (true) and (false)
                if ( ((BooleanValue*)C)->getValue() )
                {
                  if (it->isNegated()) 
                    ((ModalOp*)T)->getChild().Negate();
                  T->joinToParent(it);
                }
              }
            }
          } 
      }
    }

  // Call to the LOCAL and VIRTUAL elimination procedure... 
  // (cannot excange the order of || operands)
  ret = eliminateTrueFalse() || ret;
  
  if (checkFlag(F_ROOT))
  {
    assert(root != 0);

    // If simplifycation causes the root (this) become an unary boolean operator...
    // ...the unique child becomes the root and this node can be deleted
    if ((type == BOOLEAN_OPERATOR) && (getArity() == 1)) 
    {
      (Children.begin())->getPtr()->Parent = 0;
      (Children.begin())->getPtr()->setFlag(F_ROOT);
    
      (*root) = *(Children.begin());
    
      Children.clear();
      delete (this);
    }

    // If root is now a modal operator can be repeatly simplyficated
    eliminateModalRoot(root);
  }

  return ret;
}

void Term::eliminateModalRoot (TermPtr *root)
{
  Term *T;
  
  // While root is a modal operator, and it is not negated...
  // (note that a ~BOX or ~DIAMOND formulas could be simplyfied, ... 
  //  ...but needs to invert the polarity of the subformula, so are not handled)
  while ( ((T = root->getPtr())->getType() == MODAL_OPERATOR) && (root->isNegated() == false) )
  {
    // The following simplyfications are applied to root terms, so it shouldn't care if it is shared or not...
    assert(T->checkFlag(F_SHARED) == false);
    
    ModalOp *M = (ModalOp*)T;
    // Diamonds can be removed...
    if (M->getOperator() == O_DIAMOND)
    {
      M->getChild().getPtr()->Parent = 0;
      M->getChild().getPtr()->setFlag(F_ROOT);
    
      (*root) = M->getChild();

      T->Children.clear();
      delete (T);
    }
    // Box without diamond becomes true...
    else if (M->getOperator() == O_BOX)
    {
      BooleanValue *B = new BooleanValue(true);
      root->assign(B, false);
      
      delete (T);
    }
  }
}

bool Term::boolAssociation () 
{
  // "ret" is set to true if any of the association procedure returns true (there is a change)
  bool ret = false;

  assert(checkFlag(F_SHARED) == false);
  
  // Recursive propagation of the association procedure (BOTTOM-UP)
  for (term_it_t it = Children.begin(); it != Children.end(); ++it)
    // Cannot excange the order of || operands... 
    // otherwise "boolAssociation" could not be recursively propagated
    ret = ( it->getPtr()->boolAssociation() ) || ret;

  // Call to the LOCAL and VIRTUAL association procedure... 
  // (cannot excange the order of || operands)
  ret = flatBooleanOp() || ret;

  return ret;
}

bool Term::modalSimplifycation (TermPtr *root) 
{
  // "ret" is set to true if any modal simplifycation procedure has generated a change
  bool ret = false;

  assert(checkFlag(F_SHARED) == false);
  
  // Modal simplyfication is done top-down since it can ganerate at the level below oyher simplyfication
  // (no simplyfication can be generated at the level above if also boolean association is chosen)
  
  // Call to the LOCAL and VIRTUAL modal simplifycation procedure... 
  // (cannot excange the order of || operands)
  ret = simplifyModalOp() || ret;

  // Recursive propagation of the modal simplifycation procedure (TOP-DOWN)
  for (term_it_t it = Children.begin(); it != Children.end(); ++it)
  {
    Term *T = it->getPtr();

    // Cannot excange the order of || operands... 
    // otherwise "modalsimplifycation" could not be recursively propagated
    ret = (T->modalSimplifycation(0)) || ret;
    // If simplifycation causes a child ennary boolean operator to become unary...
    // ...it is removed and its children takes its place between the children of this node.
    if ( (T->getType() == BOOLEAN_OPERATOR) &&
         (T->getArity() == 1) )
    {
      if (it->isNegated())
        ((T->Children).begin())->Negate();
      T->joinToParent(it);
    }
  }
  
  if (checkFlag(F_ROOT))
  {
    assert(root != 0);
    
    // If simplifycation causes the root (this) become an unary boolean operator...
    // ...the unique child becomes the root and this node can be deleted
    if ((type == BOOLEAN_OPERATOR) && (getArity() == 1))
    {
      (Children.begin())->getPtr()->Parent = 0;
      (Children.begin())->getPtr()->setFlag(F_ROOT);
    
      (*root) = *(Children.begin());
    
      Children.clear();
      delete (this);
    }
    
    // If root is now a modal operator can be repeatly simplyficated
    eliminateModalRoot(root);
  }

  return ret;
}

void Term::joinToParent ()
{
  term_it_t it; 
  
  assert(checkFlag(F_SHARED) == false);
  
  it = Parent->Children.begin();
  joinToParent(it);
}

void Term::joinToParent (term_it_t &pos)
{
  term_it_t it, tmp;

  assert(checkFlag(F_SHARED) == false);
  
  // This term will be deleted, all its Children will become children of Parent
  for (it = Children.begin(); it != Children.end(); ++it)
    it->getPtr()->Parent = Parent;

  // Transfer the list of children in Parent before the iterator "pos"
  Parent->Children.splice(pos, Children);

  // "pos" is passed as reference: here it is set to a consistent value
  tmp = pos;
  --pos;
  // ...this node is removed from the Parent list and has to be deleted 
  Parent->Children.erase(tmp);
  delete (this);
  // ...in this way thew next element of "pos" is the same than before the call
}

void Term::printDotGraph (FILE *f)
{ 
  term_list_t::iterator it;
  
  // If term is in a DAG print the DOT representation only at the firs visit
  if ( checkFlag(F_SHARED) && checkFlag(F_VISITED) ) return;
  
  // Prints the DOT description of this term (a node in the graph)
  fprintf(f, "  "); 
  printDotId(f); 
  printDotAttributes(f); 
  fprintf(f, ";\n");
  
  if (checkFlag(F_SHARED)) 
    setFlag(F_VISITED);
  
  // For every child... 
  for (it = Children.begin(); it != Children.end(); ++it)
  {
    // ...print the first half of the link with the child
    fprintf(f, "\n  ");     
    printDotId(f); 
    // ...print the DOT the description of the subformula 
    it->printDotGraph(f);
  }
}  



// Constructor
BooleanValue::BooleanValue (bool v):
  Term()
{
  value = v;              // Set the boolean value
  type = BOOLEAN_VALUE;
}



// Constructor
PropositionalVar::PropositionalVar (char *n):
  Name(n),    // Create a string with the copy of the name
  Term()
{
  delete n;   // Free memory used for the variable name
  type = PROPOSITIONAL_VAR;
}

// Constructor
PropositionalVar::PropositionalVar (PropositionalVar &PV):
  Name(PV.Name),    // Create a string with the copy of the name
  Term()
{
  type = PROPOSITIONAL_VAR;
}

  

// Constructor
BooleanOp::BooleanOp (term_list_t *operands, operator_t op):
  Term(operands)  // Construtor with assignemnt of the children list
{
  bool_op = op;
  type = BOOLEAN_OPERATOR;  
}

BooleanOp::BooleanOp (TermPtr &left, TermPtr &right, operator_t op):
  Term(left, right)  // Construtor with assignemnt of the children
{
  bool_op = op;
  type = BOOLEAN_OPERATOR;  
}

void BooleanOp::pushNegationInwards (TermPtr *myPtr)
{
  if ((bool_op != O_AND) && (bool_op != O_OR))
    return;
  
  // If operator is negated
  if (myPtr->isNegated())
  {
    // Negate all children
    for (term_it_t it = Children.begin(); it != Children.end(); ++it)
      it->Negate();

    // Switch the negation of this term
    myPtr->Negate();
    
    // Swap boolean operator
    switch (bool_op)
    {
      case O_AND: bool_op = O_OR;  break;
      case O_OR:  bool_op = O_AND; break;
      default: break;
    }
  }
}

bool BooleanOp::eliminateTrueFalse ()
{
  bool modified = false;
  term_it_t it, tmp;
  
  if ((bool_op != O_AND) && (bool_op != O_OR))
    return modified;
   
  for (it = Children.begin(); it != Children.end(); ++it)
  {
    if ((it->getPtr()->getType() == BOOLEAN_VALUE)  && (getArity() > 1))
    {  
      if ( ( (BooleanValue*)(it->getPtr()) )->getValue() == (bool_op == O_AND) )
      // simplify TRUE in ANDs and FALSE in ORs... 
      // Remove and free that child
      {
        tmp = it; --tmp;
        delete (it->getPtr());
        Children.erase(it);
        it = tmp;
        modified = true;
      }
      else
      // simplify ANDs containing FALSE and ORs containing TRUE... 
      // Free all children and substitute them with the boolean child
      {
        TermPtr child = *it;
        for (tmp = Children.begin(); tmp != Children.end(); ++tmp)
          if (tmp != it)
            delete (tmp->getPtr());
        Children.clear();
          
        Children.push_front(child);
        modified = true;
      }
    }  
    if (Children.size() <= 1)
      return modified;
  }
  return modified;
}

bool BooleanOp::flatBooleanOp ()
{
  bool modified = false;
  term_it_t it, tmp;
  
  if ((bool_op != O_AND) && (bool_op != O_OR))
    return modified;
   
  // If there is a child with the same boolean operator...
  // ...joint its children to this term deleting the child
  for (it = Children.begin(); it != Children.end(); ++it)
  {
    if ( 
         ( it->getPtr()->getType() == BOOLEAN_OPERATOR) &&
         ( ((BooleanOp*)(it->getPtr()))->bool_op == bool_op) 
       )
    {
      it->getPtr()->joinToParent(it);
      modified = true;
    }
  }
  
  return modified;
}

// This method could simplify all children of the term, restricting all of them to one unique child
bool BooleanOp::simplifyModalOp ()
{
  // True if between the children of simplifycated terms there are boolean values
  bool contains_bool_value;
  // True if this procedure has modified the subformula rooted in this term
  bool modified = false;

  // Modal operator of subterms that can be simplifyed 
  operator_t simplifycable_mop; 
  // Parity of subterms that can be simplifyed 
  bool simplifycable_neg;
  // Boolean operator of the new subterm after the simplifycation
  operator_t result_op = bool_op;

  // Number of found simplifycable subterms
  int found = 0;
  // Position of the last simplyficable subterm
  term_it_t stop; 

  // Can be comprimed BOXES in an AND operator and DIAMONDS (or ~BOX) in an OR operator
  switch (bool_op)
  {  
    case O_AND: 
      simplifycable_mop = O_BOX; 
      simplifycable_neg = false; 
     // AND of BOXes  becomes  BOX of AND
      break;
    case O_OR: 
      if (Term::is_NNF)
      {
        simplifycable_mop = O_DIAMOND; 
        simplifycable_neg = false; 
        // OR of DIAMONDs  becomes  DIAMOND of OR
      }
      else if (Term::is_BNF)
      {
        simplifycable_mop = O_BOX; 
        simplifycable_neg = true; 
        // OR of ~BOXes  becomes  ~BOX of AND
        result_op = O_AND;
      }
      else return modified;
      break;
    default: 
      return modified;
  }
  
  // Swap the order of the children, pushing in front the modal children of the simplificable operator
  stop = sortModalChildren(simplifycable_mop, simplifycable_neg, &found, &contains_bool_value);

  if (found >= 2)
  {
    // Simplifications can be made only if at least two simplificable modal children are found...
    
    BooleanOp *NewBoolTerm;
    ModalOp *NewModalTerm;
    term_list_t *NewBoolChildren = new term_list_t();

    // "stop" points to the last swapped modal term...
    ++stop;    
    // ...now "stop" points to the first of the non-swapped terms
    
    // Join all found modal children... so they dissappear and their respective children becomes children of this
    for (term_it_t it = Children.begin(); it != stop; ++it)
      it->getPtr()->joinToParent(it);
  
    // Move these new children in a new list that represents the children of a new boolean term
    NewBoolChildren->splice(NewBoolChildren->begin(), Children, Children.begin(), stop);
    NewBoolTerm = new BooleanOp(NewBoolChildren, result_op);

    // If the new boolean term contains true or false it has to be simplifycated
    if (contains_bool_value)
      NewBoolTerm->eliminateTrueFalse();
    
    // The new boolean operator becomes the child of new modal term with the simplifycable operator 
    TermPtr Tmp((Term *)NewBoolTerm, false);
    NewModalTerm = new ModalOp(Tmp, simplifycable_mop);
    
    // The new modal term become child of this term
    Tmp.assign((Term *)NewModalTerm, simplifycable_neg); 
    addChild(Tmp); 

    // If simplifycation of true and false leads to an unary boolean operator it can be removed...  
    if (contains_bool_value && (NewBoolTerm->getArity() == 1))
      NewBoolTerm->joinToParent();
    
    modified = true;
  }

  return modified;
}

term_it_t BooleanOp::sortModalChildren (operator_t modal_op, bool polarity, int *swapped, bool *contains_boolean)
{
  int i = 0;
  
  *contains_boolean = false;
  
  term_it_t last_of_swapped = Children.end(); 
  term_it_t it = Children.begin();

  while (it != Children.end()) 
  {
    if ( ( it->isNegated() == polarity ) && 
         ( it->getPtr()->getType() == MODAL_OPERATOR ) &&
         ( ( (ModalOp*)(it->getPtr()) )->getOperator() == modal_op ) )
    {
      if ( ((ModalOp*)it->getPtr())->getChild().getPtr()->getType() == BOOLEAN_VALUE )
        *contains_boolean = true;
      
      Children.push_front(*it);
      it = Children.erase(it);
      ++i;
      if (i == 1) 
        last_of_swapped = Children.begin();
    }
    else
      ++it;
  }
  
  *swapped = i;
  return last_of_swapped;
}



ModalOp::ModalOp (TermPtr &sub_formula, operator_t op):
  Term(sub_formula)
{
  modal_op = op;  
  type = MODAL_OPERATOR;  
}

void ModalOp::pushNegationInwards (TermPtr *myPtr)
{
  if ((modal_op != O_BOX) && (modal_op != O_DIAMOND))
    return;

  if (myPtr->isNegated())
  {
    // Negate the child
    (Children.begin())->Negate();
    
    // Switch the negation of this term
    myPtr->Negate();

    // Swap modal operator
    switch (modal_op)
    {
      case O_BOX: modal_op = O_DIAMOND; break;
      case O_DIAMOND: modal_op = O_BOX; break;
      default: break;
    }
  }
}

void ModalOp::toBoxNormalForm (TermPtr *myPtr)
{
  if (modal_op != O_DIAMOND)
    return;

  // Here modal_op is surely DIAMOND...

  // ...DIAMOND(x) is equivalent to not(BOX(not(x)))
  // Negate the child
  (Children.begin())->Negate();
  // Switch the negation of this term
  myPtr->Negate();
  // Swap to BOX operator
  modal_op = O_BOX;
}



// Generic method for all term types
unsigned long Term::getCnfId (Label *l, unsigned long use_id, FILE *map_f, transl_t translation)
{
  if (use_id || (last_transl_label_id != l->getId())) 
  { 
    if (use_id)
      // The CNF id value is forced 
      last_transl_id = use_id;
    else
      // Generate a new CNF id since the label is different with respect to the last used
      last_transl_id = ++Term::cnf_id_counter;
    last_transl_label_id = l->getId();
    
    // The term has to be re-translate in the new label...
    unsetFlag(F_TRANSLATED_P | F_TRANSLATED_N);
    
    // If is a propositiona variable and is open a file, print the mapping between the id, the label, and the variable name
    if ((type == PROPOSITIONAL_VAR) && map_f)
      ((PropositionalVar *)this)->printVarMap(map_f, translation, l);
  }

  // Return the current CNF id  
  return last_transl_id;
}

// Method only for boolean value ter
unsigned long BooleanValue::getCnfId (Label *l, unsigned long use_id, FILE *map_f, transl_t translation)
{
  // If a CNF id have never been assigned ...
  if (last_transl_id == 0)
  {
    // Generate a new CNF id (that will never change anymore...)
    last_transl_id = ++Term::cnf_id_counter;
    last_transl_label_id = 0;

    unsetFlag(F_TRANSLATED_P | F_TRANSLATED_N);
    
    // Print the mapping between the id and the boolean value
    printVarMap(map_f, translation, l);
    
    // Note: translation of boolean values can not be done here...
    // ...since otherwise it will be printed in the middle of the translation of another term
  }
  // ...otherwise the id is always the same, indipendently from the label
  
  return last_transl_id;
}

int Term::translateAsRoot (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, Label *l, bool negated)
{
  // The root has to be satisfiable...
  if (negated) fprintf(out_f, "-");
  fprintf(out_f, "%d 0", getCnfId(l));
  
  return 1;
}

int Term::translateNonModal (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l, bool negated)
{ return 0; }

int BooleanOp::translateNonModal (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l, bool negated)
{
  int clauses = 0;
  
  // Every term have to be translated once (in NNF)
  // (in BNF every term can be translated twice: one positive and one negative)
  if ( ((!negated) && checkFlag(F_TRANSLATED_P)) ||
       (negated && checkFlag(F_TRANSLATED_N)) )
    return clauses;
    
  // Print the translation of the boolean term... 
  clauses = translateBoolOp(out_f, header_f, map_f, translation, l, negated);

  // Propagate translation to children
  for (term_it_t c = Children.begin(); c != Children.end(); ++c)
    clauses += c->getPtr()->translateNonModal(out_f, header_f, map_f, translation, Diamonds, Boxes, l, (c->isNegated() ^ negated) /*XOR*/); 
  
  return clauses;
}

int ModalOp::translateNonModal (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l, bool negated)
{
  term_transl_struct T;

  if ( ((!negated) && checkFlag(F_TRANSLATED_P)) ||
       (  negated  && checkFlag(F_TRANSLATED_N)) )
    return 0;
  // This assures that the term is insert in the respective list only once...
  // (In BNF the same BOX term could be inserted once in both lists, depending on the polarity N or P)
  
  T.term = this;
  T.negated = negated;  
  T.cnf_id = getCnfId(l);
  
  switch (modal_op)
  {
    case O_BOX: 
      if (Term::is_BNF && negated)
      {
        // In BNF a ~BOX (a box occouring negated or with negative polarity) is handled as a Diamond...
        // ...since it generates new state in translation
        Diamonds->push_back(T); 
        setFlag(F_TRANSLATED_N); 
      }
      else
      {
        Boxes->push_back(T); 
        setFlag(F_TRANSLATED_P); 
      }
      break;
    case O_DIAMOND: 
      Diamonds->push_back(T); 
      setFlag(F_TRANSLATED_P); 
      break;
  }
  
  return 0;
}

/*
int PropositionalVar::translateNonModal (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, list<term_transl_struct> *Diamonds, list<term_transl_struct> *Boxes, Label *l, bool negated)
{
  if (checkFlag(F_TRANSLATED_P | F_TRANSLATED_N)) return 0;
  printVarMap(map_f, translation, l);
  return 0;
}
*/

int BooleanValue::printVarMap (FILE *map_f, transl_t translation, Label *l)
{
  if (map_f) 
  {
    // Print CNF ID, boolean value, <empty> (since the id is unique for every label)
    fprintf(map_f, "%d\t%s\t\n", last_transl_id, (value ? "true" : "false") );
    return 1;
  }
  else
    return 0;  
}

int PropositionalVar::printVarMap (FILE *map_f, transl_t translation, Label *l)
{
  setFlag(F_TRANSLATED_P | F_TRANSLATED_N);
  if (map_f) 
  {
    // Print CNF ID, variable name, label (state)
    fprintf(map_f, "%d\t%s\t%s\n", getCnfId(l), Name.c_str(), l->getText());
    return 1;
  }
  else
    return 0;  
}


int BooleanOp::translateBoolOp (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, Label *l, bool negated)
{
  bool is_and_op = ((bool_op == O_AND) ^ negated);
  int clauses; 
  
    
  if (is_and_op)
  {
    // AND operator: " this ->   AND_i ( child_i ) "... in CNF: "AND_i ( this ->  child_i)"
    // -OR operator: "-this -> -( OR_i ( child_i ))"... in CNF: "AND_i (-this -> -child_i)"
    // (It requires Children.size() clauses, each one of two literals)

    for (term_it_t c = Children.begin(); c != Children.end(); ++c)
    {
      // "t -> c_i" is "-t OR c_i" 
      fprintf(out_f, "\n");
      if (! negated)
        fprintf(out_f, "-");
      fprintf(out_f, "%d ", getCnfId(l));
      if ( c->isNegated() ^ negated )
        fprintf(out_f, "-");
      fprintf(out_f, "%d 0", c->getPtr()->getCnfId(l, 0, map_f, translation));
    }

    clauses = Children.size();
  }
  
  if (!is_and_op)
  {
    //   OR operator: " this ->    OR_i ( child_i ) "... the same in CNF
    // -AND operator: "-this -> -(AND_i ( child_i ))"... in CNF: "-this -> OR_i ( -child_i )"
    // (It requires one clause of Children.size()+1 literals)

    // "t -> OR_i c_i" is "-t OR (OR_i c_i)" 
    fprintf(out_f, "\n");
    if (! negated)
      fprintf(out_f, "-");
    fprintf(out_f, "%d", getCnfId(l)); 
    for (term_it_t c = Children.begin(); c != Children.end(); ++c) 
    {
      fprintf(out_f, " ");
      if ( c->isNegated() ^ negated )
        fprintf(out_f, "-");
      fprintf(out_f, "%d", c->getPtr()->getCnfId(l, 0, map_f, translation));
    }
    fprintf(out_f, " 0");

    clauses = 1;
  }
  
  if (negated) 
    setFlag(F_TRANSLATED_N);
  else
    setFlag(F_TRANSLATED_P);
  return clauses;
}

int ModalOp::translateBox (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, Label *l, unsigned long use_id, Label *child_l, unsigned long id_diamond)
{
  int clauses = 1;
  Term *C = getChild().getPtr();

  
  if (Term::is_BNF && (use_id == id_diamond))
    // [BNF] Translation of a BOX operator with respect to the same negated BOX operator is not possible...
    // ... in fact only one of the two subformulas can be satisfiable (true)
    return 0;
  
  // [NNF] Translation of a BOX operator: "(in the current state) BOX_id & DIAMOND_id -> (in a new child state) CHILD_id"
  // [BNF] Translation of a BOX operator: "(in the current state) BOX_id & ~ BOX_id_2 -> (in a new child state) CHILD_id"
  
  if (! child_l) 
    // There is no a child label, so there no exists diamonds in the same state of the box
    fprintf(out_f, "\n%d 0", use_id); // In K the box (without dimaonds) is always true
  else
  {
    fprintf(out_f, "\n-%d ", use_id); 
    if (Term::is_NNF)        
      fprintf(out_f, "-");             // [NNF] Since "(B  &   D) -> C" in CNF is "-B  v -D v C"
    fprintf(out_f, "%d ", id_diamond); // [BNF] Since "(B1 & -B2) -> C" in CNF is "-B1 v B2 v C"
    if (getChild().isNegated())
      fprintf(out_f, "-");
    fprintf(out_f, "%d 0", C->getCnfId(child_l, 0, map_f, translation));
    
    // Boolean values can be child only of modal terms since are eliminate from boolean-op term...
    // ...so translation has to be called here
    if (C->getType() == BOOLEAN_VALUE) 
      clauses += ((BooleanValue*)C)->translateBooleanValue (out_f, header_f, map_f, translation);
  }
  
  return clauses;
}

int ModalOp::translateDiamond (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation, Label *l, unsigned long use_id, Label *child_l)
{
  int clauses = 1;
  Term *C = getChild().getPtr();
  
  
  if (Term::is_NNF)
  {
    // Translation of a DIAMOND operator: "(in the current state) DIAMOND_id -> (in a new child state) CHILD_id" 
    fprintf(out_f, "\n-%d ", use_id);
    if (getChild().isNegated())
      fprintf(out_f, "-");
    fprintf(out_f, "%d 0", C->getCnfId(child_l, 0, map_f, translation));
  }
  else if (Term::is_BNF)
  {
    // Translation of a ~ BOX operator: "(in the current state) ~ BOX_id -> (in a new child state) ~ CHILD_id"
    fprintf(out_f, "\n%d ", use_id); // Since "-B -> -C" ...in CNF is: "B v -C"
    if (! getChild().isNegated())    // Diamond is a negated box, so child has to be negated
      fprintf(out_f, "-");
    fprintf(out_f, "%d 0", C->getCnfId(child_l, 0, map_f, translation));
  }

  // Boolean values can be child only of modal terms since are eliminate from boolean-op term...
  // ...so translation has to be called here
  if (C->getType() == BOOLEAN_VALUE) 
    clauses += ((BooleanValue*)C)->translateBooleanValue (out_f, header_f, map_f, translation);
  
  return clauses;
}

int BooleanValue::translateBooleanValue (FILE *out_f, FILE *header_f, FILE *map_f, transl_t translation)
{
  if (checkFlag(F_TRANSLATED_P | F_TRANSLATED_N)) return 0;
  setFlag(F_TRANSLATED_P | F_TRANSLATED_N);
  
  fprintf(out_f, "\n%s%d 0", (value ? "" : "-"), last_transl_id);
  return 1;  
}
